"""
.. module:: Katna.image_filters.text_detector
    :platform: OS X
    :synopsis: This module is implementation of text detector filter
"""

import os
import cv2
import numpy as np
import time
import requests
import random
from imutils.object_detection import non_max_suppression
from Katna.image_filters.filter import Filter
import Katna.config as config


class TextDetector(Filter):
    """TextDetector Class: Class for implementation of text detector filter, inherit from Filter class
    """

    def __init__(self, weight=1.0):
        """Constructor for this class does following tasks, if not already downloaded\
        , it first downloads text detector dnn weights file from public URL\
        ands save it at USER_HOME/.katna directory, or /tmp/.katna directory.\
        After this initializer code initializes internal parameter: \
        min_confidence (for text detection)
        """
        super().__init__(weight)
        self.min_confidence = config.TextDetector.min_confidence
        self.merge_threshold = config.TextDetector.merge_threshold
        self.layerNames = config.TextDetector.layerNames
        self.frozen_weights = config.TextDetector.frozen_weights
        self.cache_subdir = config.TextDetector.cache_subdir

        try:
            self.network_folder_path = os.path.join(os.path.expanduser("~"), ".katna")
            if not os.access(self.network_folder_path, os.W_OK):
                self.network_folder_path = os.path.join("/tmp", ".katna")
            self.datadir = os.path.join(self.network_folder_path, self.cache_subdir)
            if not os.path.exists(self.datadir):
                os.makedirs(self.datadir)

            self.network_file_path = os.path.join(self.datadir, self.frozen_weights)
            if not os.path.exists(self.network_file_path):
                self.download_data()

            self.net = cv2.dnn.readNet(self.network_file_path)

        except Exception:
            raise FileNotFoundError(
                self.frozen_weights
                + " seems to be missing.\
                 Download the file and specify the full path\
                      while initializing TextDetector class"
            )

    def download_data(self):
        """Public function for downloading the network weight from the URL link, to be used for
        text detection functionality. 
        Troubleshooting tip: If you get FileNotFound error during text detector initialization,
        initialize the text detector and call this function directly to download the model file from public URL link.
        """
        # create response object
        link = config.TextDetector.model_download_link
        r = requests.get(link, stream=True)
        # download started
        print("Downloading model file...")
        # if not os.path.isfile(self.network_file_path) or not os.path.exists(self.network_file_path):
        with open(os.path.join(self.datadir, self.frozen_weights), "wb") as f:
            for chunk in r.iter_content(chunk_size=1024 * 1024):
                if chunk:
                    f.write(chunk)
        print("Model file downloaded.")

    def __decode_predictions(self, scores, geometry):
        """Internal Function for getting bounding box and confidence values 
        from text detector dnn network output (scores, geometry)
        function takes the number of rows and columns from the scores volume, then
        initializes set of bounding box rectangles and corresponding confidence scores
        """
        (numRows, numCols) = scores.shape[2:4]
        rects = []
        confidences = []

        # loop over the number of rows
        for y in range(0, numRows):
            # extract the scores (probabilities), followed by the
            # geometrical data used to derive potential bounding box
            # coordinates that surround text
            scoresData = scores[0, 0, y]
            xData0 = geometry[0, 0, y]
            xData1 = geometry[0, 1, y]
            xData2 = geometry[0, 2, y]
            xData3 = geometry[0, 3, y]
            anglesData = geometry[0, 4, y]

            # loop over the number of columns
            for x in range(0, numCols):
                # if our score does not have sufficient probability,
                # ignore it
                if scoresData[x] < self.min_confidence:
                    continue

                # compute the offset factor as our resulting feature
                # maps will be 4x smaller than the input image
                (offsetX, offsetY) = (x * 4.0, y * 4.0)

                # extract the rotation angle for the prediction and
                # then compute the sin and cosine
                angle = anglesData[x]
                cos = np.cos(angle)
                sin = np.sin(angle)

                # use the geometry volume to derive the width and height
                # of the bounding box
                h = xData0[x] + xData2[x]
                w = xData1[x] + xData3[x]

                # compute both the starting and ending (x, y)-coordinates
                # for the text prediction bounding box
                endX = int(offsetX + (cos * xData1[x]) + (sin * xData2[x]))
                endY = int(offsetY - (sin * xData1[x]) + (cos * xData2[x]))
                startX = int(endX - w)
                startY = int(endY - h)

                # add the bounding box coordinates and probability score
                # to our respective lists
                rects.append((startX, startY, endX, endY))
                confidences.append(scoresData[x])

        # return a tuple of the bounding boxes and associated confidences
        return (rects, confidences)

    def __merge_boxes(self, rects):
        """main function to detect text boxes from image

        :param rects: list of 
        :type rects: numpy array
        :param rectsUsed: image file in numpy array/opencv format
        :type rectsUsed: numpy array

        :return: output image with the list of text boxes
        :rtype: file, list
        """

        def grouper(iterable, interval=2):
            prev = None
            group = []
            for item in iterable:
                if not prev or abs(item[1] - prev[1]) <= interval:
                    group.append(item)
                else:
                    yield group
                    group = [item]
                prev = item
            if group:
                yield group

        rects_used = []
        heights = list()
        for bbox in rects:
            heights.append(bbox[3] - bbox[1])
        heights = sorted(heights)  # Sort heights
        median_height = heights[len(heights) // 2] / 2  # Find half of the median height

        bboxes_list = sorted(
            rects, key=lambda k: k[1]
        )  # Sort the bounding boxes based on y1 coordinate ( y of the left-top coordinate )
        combined_bboxes = grouper(
            bboxes_list, median_height
        )  # Group the bounding boxes
        for group in combined_bboxes:
            x_min = min(group, key=lambda k: k[0])[0]  # Find min of x1
            x_max = max(group, key=lambda k: k[2])[2]  # Find max of x2
            y_min = min(group, key=lambda k: k[1])[1]  # Find min of y1
            y_max = max(group, key=lambda k: k[3])[3]  # Find max of y2
            rects_used.append([x_min, y_min, x_max, y_max])
        return rects_used

    def __detect_text(self):
        """Internal function to detect text bounding boxes from input image.
        Returns list of bounding boxes of each detected text field in input image.

        :param image: image file in numpy array/opencv format
        :type image: numpy array
        :param output_image: image file in numpy array/opencv format
        :type output_image: numpy array

        :return: output image with the list of text boxes
        :rtype: file, list
        """
        (H, W) = self.image.shape[:2]
        rW = W / 320
        rH = H / 320
        image = cv2.resize(self.image, (320, 320))
        (H, W) = image.shape[:2]

        # construct a blob from the image and then perform a forward pass of
        # the model to obtain the two output layer sets
        blob = cv2.dnn.blobFromImage(
            self.image, 1.0, (W, H), (123.68, 116.78, 103.94), swapRB=True, crop=False
        )

        self.net.setInput(blob)
        (scores, geometry) = self.net.forward(self.layerNames)

        rects, confidences = self.__decode_predictions(scores, geometry)
        # apply non-maxima suppression to suppress weak, overlapping bounding
        # boxes
        boxes = non_max_suppression(np.array(rects), probs=confidences)
        text_rects = []
        # loop over the bounding boxes
        for (startX, startY, endX, endY) in boxes:
            # scale the bounding box coordinates based on the respective
            # ratios

            startX = int(startX * rW)
            startY = int(startY * rH)
            endX = int(endX * rW)
            endY = int(endY * rH)
            cv2.rectangle(self.image, (startX, startY), (endX, endY), (0, 0, 255), 3)
            text_rects.append([startX, startY, endX, endY])

        text_rects = sorted(text_rects, key=lambda item: item[0])
        final_rects = text_rects
        if len(text_rects) > 0:
            final_rects = self.__merge_boxes(text_rects)

        return final_rects

    def set_image(self, image):
        """Public set_image function, This will detect all text boxes in input image and
        will saves them as internal list of text_rect to be used in get_filter_result

        :param image: input image from which needs to be cropped
        :type image: numpy array(opencv)
        """
        if image is None:
            return None
        self.image = image
        self.text_rects = self.__detect_text()

    def get_filter_result(self, crop):
        """Main public function of TextDetector filter class,
        this filter Returns false if crop contains no text, additionally
        checks for overlap between input crop rectangle and the detected
        text bounding box, returns True if No overlap (Filter will not discard input crop)
        otherwise returns False (signal for discarding input crop).
        
        :param crop: input crop rectangle to test
        :type crop: crop_rect
        :return: True if No overlap (Filter will not discard input crop) otherwise returns False 
        :rtype: bool
        """
        # rect: xs,ys,xe,ye
        # crop: x,y,w,h
        if self.text_rects is None or len(self.text_rects) == 0:
            return True

        for rect in self.text_rects:
            if not (
                (rect[2]) <= (crop.x + crop.w)
                and (rect[0]) >= (crop.x)
                and (rect[1]) >= (crop.y)
                and (rect[3]) <= (crop.y + crop.h)
            ):
                return False
            else:
                return True

        return True



"""
.. module:: Katna.image_features.edge_feature
    :platform: OS X
    :synopsis: This module is for for getting Edge feature Map from input image
"""

import os
import cv2
import numpy as np
from Katna.image_features.feature import Feature
import Katna.config as config


class EdgeFeature(Feature):
    """Class for getting edge detector Feature,
    Constructor Parameters:- feature weight \n 
    Internal Parameters:- \n 
    min_val_threshold : min edge threshold \n 
    max_val_threshold : max edge threshold \n 
    ksize: size of Sobel kernel used for finding image gradients
    """

    def __init__(self, weight=0.0):
        super().__init__(weight)
        # min edge threshold value
        self.min_val_threshold = config.EdgeFeature.min_val_threshold

        # Max edge threshold value
        self.max_val_threshold = config.EdgeFeature.max_val_threshold

        # aperture_size/size of Sobel kernel for canny edge detector
        self.ksize = config.EdgeFeature.ksize

    def get_feature_map(self, image):
        """Public function for getting Feature map image from edges detection

        :param image: input image
        :type image: `numpy array`
        :return: single channel opencv numpy image with feature map from edge detection
        :rtype: numpy array
        """
        return cv2.Canny(
            image, self.min_val_threshold, self.max_val_threshold, self.ksize
        )



"""
.. module:: Katna.crop_rect
    :platform: OS X
    :synopsis: This module defines crop spec
"""
import os
import cv2
import numpy as np
import Katna.config as config


class CropRect(object):
    """Data structure class for storing image crop rectangles

    :param object: base class inheritance
    :type object: class:`Object`
    """

    def __init__(self, x, y, w, h):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.score = 0.0
        self.debug_image = 0

    def __str__(self):
        """Print crop rectangle

        :param object: base class inheritance
        :type object: class:`Object`
        """
        rep = (
            " x_pos: "
            + str(self.x)
            + " y_pos: "
            + str(self.y)
            + " width: "
            + str(self.w)
            + " height: "
            + str(self.h)
        )
        return rep

    def get_image_crop(self, input_image):
        """public functions which for given input image and current crop rectangle object returns image cropped to 
        crop rectangle specifications. 

        :param object: base class inheritance
        :type object: class:`Object`
        :param image: input image
        :type image: Opencv Numpy Image
        :return: cropped image according to given spec
        :rtype: Opencv Numpy Image
        """
        crop_img = input_image[self.y : self.y + self.h, self.x : self.x + self.w]
        return crop_img



"""
.. module:: tests.image_similarity
    :platform: OS X
    :synopsis: This module is used for comparsion of image 
"""

import warnings
from skimage.measure import compare_ssim
from skimage.transform import resize
from scipy.stats import wasserstein_distance
import scipy
import skimage
import numpy as np
import cv2
import imageio


class ImageSimilarity(object):
    """Class for performing Image similairty comparision between images

  :param object: None
  :type object:
  """

    def __init__(self):

        # specify resized image sizes
        self.height = 2 ** 7
        self.width = 2 ** 7

    def normalize_exposure(self, img):
        """
    Normalize the exposure of an image.
    param img: the input image
    type x: numpy.ndarray
    return: the normalized image
    rtype: numpy.ndarray
    """
        img = img.astype(int)
        hist = self.get_histogram(img)
        # get the sum of vals accumulated by each position in hist
        cdf = np.array([sum(hist[: i + 1]) for i in range(len(hist))])
        # determine the normalization values for each unit of the cdf
        sk = np.uint8(255 * cdf)
        # normalize each position in the output image
        height, width = img.shape
        normalized = np.zeros_like(img)
        for i in range(0, height):
            for j in range(0, width):
                normalized[i, j] = sk[img[i, j]]
        return normalized.astype(int)

    def get_histogram(self, img):
        """
    Get the histogram of an image. For an 8-bit, grayscale image, the
    histogram will be a 256 unit vector in which the nth value indicates
    the percent of the pixels in the image with the given darkness level.
    The histogram's values sum to 1.
    param img: the input image
    type x: numpy.ndarray
    return: the image histogram 
    rtype: numpy.ndarray
    """
        h, w = img.shape
        hist = [0.0] * 256
        for i in range(h):
            for j in range(w):
                hist[img[i, j]] += 1
        return np.array(hist) / (h * w)

    def get_img(self, path, norm_size=True, norm_exposure=False):
        """
    Prepare an image for image processing tasks
    param path: the input image path
    type x: str
    return: the image
    rtype: numpy.ndarray
    """
        # flatten returns a 2d grayscale array
        img = imageio.imread(path, as_gray=True).astype(int)
        # resizing returns float vals 0:255; convert to ints for downstream tasks
        if norm_size:
            img = skimage.transform.resize(
                img, (self.height, self.width), mode="constant", preserve_range=True
            )
        if norm_exposure:
            img = self.normalize_exposure(img)
        return img

    def pixel_sim(self, path_a, path_b):
        """
    Measure the pixel-level similarity between two images
    param path_a: the path to an image file
    type path_a: str
    param path_b: the path to an image file
    type path_b: str
    return: a float {-1:1} that measures structural similaritybetween the input images
    rtype: {float}
    """
        img_a = self.get_img(path_a, norm_exposure=True)
        img_b = self.get_img(path_b, norm_exposure=True)
        return np.sum(np.absolute(img_a - img_b)) / (self.height * self.width) / 255

    def earth_movers_distance(self, path_a, path_b):
        """
    Measure the Earth Mover's distance between two images
    type path_a: str
    param path_b: the path to an image file
    type path_b: str
    return: a float {-1:1} that measures structural similarity between the input images
    rtype: {float}
    """
        img_a = self.get_img(path_a, norm_exposure=True)
        img_b = self.get_img(path_b, norm_exposure=True)
        hist_a = self.get_histogram(img_a)
        hist_b = self.get_histogram(img_b)
        return wasserstein_distance(hist_a, hist_b)

    def structural_sim(self, path_a, path_b):
        """
    Measure the structural similarity between two images
    type path_a: str
    param path_b: the path to an image file
    type path_b: str
    return: a float {-1:1} that measures structural similarity between the input images
    rtype: {float}
    """
        img_a = self.get_img(path_a)
        img_b = self.get_img(path_b)
        sim, diff = compare_ssim(img_a, img_b, full=True)
        return sim

    def sift_sim(self, path_a, path_b):
        """
    Use SIFT features to measure image similarity
    type path_a: str
    param path_b: the path to an image file
    type path_b: str
    return: a float {-1:1} that measures  similarity between the input images
    rtype: {float}
    """
        # initialize the sift feature detector
        orb = cv2.ORB_create()

        # get the images
        img_a = cv2.imread(path_a)
        img_b = cv2.imread(path_b)

        # find the keypoints and descriptors with SIFT
        kp_a, desc_a = orb.detectAndCompute(img_a, None)
        kp_b, desc_b = orb.detectAndCompute(img_b, None)

        # initialize the bruteforce matcher
        bf = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)

        # match.distance is a float between {0:100} - lower means more similar
        matches = bf.match(desc_a, desc_b)
        similar_regions = [i for i in matches if i.distance < 70]
        if len(matches) == 0:
            return 0
        return len(similar_regions) / len(matches)



"""
.. module:: Katna.crop_extractor
    :platform: OS X
    :synopsis: This module has functions related to crop extraction
"""

import cv2
import numpy as np
import sys
from Katna.crop_rect import CropRect
import math
import time
import Katna.config as config
from Katna.decorators import DebugDecorators


class CropExtractor(object):
    """Class for extraction and scoring of all possible crops from input image for input crop size.
    Currently features being used for scoring a given crop retangle are: Edge, saliency and face detection.
    Additionally crop scoring also takes following metrics into account: Distance of pixel in crop rectangle from crop boundary
    and rule of third.
    """

    def __init__(self):

        self.detail_weight = config.CropScorer.detail_weight
        self.edge_radius = config.CropScorer.edge_radius
        self.edge_weight = config.CropScorer.edge_weight
        self.outside_importance = config.CropScorer.outside_importance
        self.rule_of_thirds = config.CropScorer.rule_of_thirds
        self.saliency_bias = config.CropScorer.saliency_bias
        self.saliency_weight = config.CropScorer.saliency_weight
        self.down_sample_factor = config.Image.down_sample_factor
        self.face_bias = config.CropScorer.face_bias
        self.face_weight = config.CropScorer.face_weight
        self.rects_weight = config.CropScorer.rects_weight
        self.extracted_feature_maps = []

    def _get_all_possible_crops(
        self,
        image,
        crop_width,
        crop_height,
        max_scale=1,
        min_scale=0.9,
        scale_step=0.1,
        step=8,
    ):
        """Internal function for getting list of all possible Crop_rect for input image.

        :param object: base class inheritance
        :type object: class:`Object`
        :param image: image file
        :type image: OpenCV numpy Image file
        :param crop_width: width
        :type crop_width: int (pixels)
        :param crop_height: height
        :type crop_height: int (pixels)
        :param max_scale: maximum scale factor
        :type max_scale: int, default=1
        :param min_scale: minimum scale factor
        :type min_scale: int, default=0.9
        :param scale_step: scaling step
        :type scale_step: int, default=0.1
        :param step: step
        :type step: int, default=8
        :return: all_possible_crops_rect
        :rtype: list of CropRect
        """
        # This function takes a sliding window approach for getting all possible crop 
        # with given crop specification. 
        image_width, image_height = image.shape[1], image.shape[0]
        # print("Width,height", image_width, image_height)
        # get all possible crops using sliding window

        # Variable to collect all possible crops 
        all_possible_crops_rects = []

        # Start sliding window from from range 0 to max_scale
        for scale in (
            i / 100
            for i in range(
                int(max_scale * 100),
                int((min_scale - scale_step) * 100),
                -int(scale_step * 100),
            )
        ):
            for y in range(0, image_height, step):
                if not (y + crop_height * scale) <= image_height:
                    break
                for x in range(0, image_width, step):
                    if not (x + crop_width * scale) <= image_width:
                        break
                    all_possible_crops_rects.append(
                        CropRect(x, y, crop_width * scale, crop_height * scale)
                    )
        if not all_possible_crops_rects:
            raise ValueError(locals())
        return all_possible_crops_rects

    def _thirds(self, x):
        """Checks rule of third for particular x. Gets value in the range of [0, 1] where 0 is the center of the pictures
        returns weight of rule of thirds [0, 1]

        :param x: parameter x for which Rule of third needs to be checked
        :type x: int
        :return: weight of rule of thirds [0, 1]
        :rtype: int
        """
        # Rule 
        x = ((x + 2 / 3) % 2 * 0.5 - 0.5) * 16
        return np.maximum(1 - x * x, 0)

    def _importance(self, crop_rect, importance_map):
        """Function to Calculate importance_map of crop rectangle according to following criteria, Crop rectangle distance from edge
        and Rule of third

        :param object: base class inheritance
        :type object: class:`Object`
        :param crop: crop window
        :type crop_rect: crop_rect data structure
        :param importance_map: numpy array of size equal to input image and intialized with outside_importance parameter(default=-0.5)
        :type importance_map: numpy array
        :return: modified importance array
        :rtype: numpy array
        """
        # Select image around a grid size of crop width and crop height around 
        # crop center
        y, x = np.ogrid[
            int(crop_rect.y) : int(crop_rect.y + crop_rect.h),
            int(crop_rect.x) : int(crop_rect.x + crop_rect.w),
        ]

        # Subtract values in grid with Crop retangle center value so
        # Values far from center of grid would have hight amplitude
        x, y = (x - crop_rect.x) / crop_rect.w, (y - crop_rect.y) / crop_rect.h

        # Apply abs operation around center of grid 
        px, py = abs(0.5 - x) * 2, abs(0.5 - y) * 2
        px1 = px - 1 + self.edge_radius
        py1 = py - 1 + self.edge_radius
        dx = np.maximum(px1, np.zeros(px1.shape))
        dy = np.maximum(py1, np.zeros(py1.shape))

        # calculating edge distance
        d = (dx * dx + dy * dy) * self.edge_weight
        s = 1.41 - np.sqrt(px * px + py * py)

        # check for rule of thirds
        if self.rule_of_thirds:
            s = (np.maximum(0, s + d + 0.5) * 1.2) * (
                self._thirds(px) + self._thirds(py)
            )
        sdSum = s + d

        # Updating the value of importance matrix using rule of thirds and edge distance
        importance_map[
            int(crop_rect.y) : int(crop_rect.y + crop_rect.h),
            int(crop_rect.x) : int(crop_rect.x + crop_rect.w),
        ] = sdSum
        # should return wxh/hxw
        return importance_map

    def _score(self, feature_maps_image, crop_rect):
        """Internal function to get score for given crop rectangle - it calculates the importance of each pixel inside image
        for given crop_rect and calculates the score by combining all the image features with importance

        :param object: base class inheritance
        :type object: class:`Object`
        :param feature_maps_image: feature_maps image
        :type feature_maps_image: OpenCV image object
        :param crop_rect: Input crop rectangle for which score is calculated
        :type crop_rect: crop_rect
        :return: score
        :rtype: float
        """
        # setting the default scores for each crop rectangle
        rect_score_detail = 0.0
        rect_score_saliency = 0.0
        rect_score_face = 0.0
        rect_score_rects = 0.0
        rect_score_total = 0.0

        feature_maps_data = feature_maps_image
        feature_maps_width, feature_maps_height = (
            feature_maps_image.shape[1],
            feature_maps_image.shape[0],
        )
        # # Skin , Edge , Saliency are at index 0, 1, 2 respectively
        # dummy importance matrix to store base values which will be updated in _importance function
        importance_map = np.zeros((feature_maps_height, feature_maps_width))
        importance_map[:, :] = self.outside_importance
        # importance score matrix wrt rule of thirds and edge radius
        importance = self._importance(crop_rect, importance_map)

        if config.Image.DEBUG is True:
            crop_rect.debug_image = importance.copy()
            importance_max = np.max(crop_rect.debug_image)
            importance_min = np.min(crop_rect.debug_image)
            float_range = importance_max - importance_min
            crop_rect.debug_image = (
                crop_rect.debug_image - importance_min
            ) / float_range
            crop_rect.debug_image = 255 * crop_rect.debug_image
            crop_rect.debug_image = crop_rect.debug_image.astype(np.uint8)

        # print("Debug Image",crop_rect.debug_image)
        detail = np.array(feature_maps_data)[:, :, 2] / 255
        # detailT = detail.T
        rect_score_rects = np.sum(rect_score_rects * importance)
        # Skin score calculations
        a = (np.array(feature_maps_data)[:, :, 1] / 255) * (detail + self.face_bias)
        b = a * importance
        rect_score_face = np.sum(b)
        # edge score calculations
        rect_score_detail = detail * importance
        rect_score_detail = np.sum(rect_score_detail)
        # Saliency score calculations
        rect_score_saliency = np.sum(
            (np.array(feature_maps_data)[:, :, 0] / 255)
            * (detail + self.saliency_bias)
            * importance
        )

        # Scoring mechanism
        rect_score_total = (
            rect_score_detail * self.detail_weight
            + rect_score_face * self.face_weight
            + rect_score_rects * self.rects_weight
            + rect_score_saliency * self.saliency_weight
        ) / (crop_rect.w * crop_rect.h)

        return rect_score_total

    def _extract_and_score_crop_rects(
        self,
        image,
        extracted_feature_maps,
        feature_names,
        crop_width,
        crop_height,
        max_scale=1,
        min_scale=0.9,
        scale_step=0.1,
        step=8,
    ):
        """Internal function to get all crop rects and score each crop retangle given input feature maps for image

            :param object: base class inheritance
            :type object: class:`Object`
            :param image: image file
            :type image: OpenCV numpy Image file
            :param extracted_feature_maps: feature maps
            :type extracted_feature_maps: OpenCV numpy Image file
            :param feature_names: names of features used
            :type feature_names: list
            :param crop_width: width
            :type crop_width: int (pixels)
            :param crop_height: height
            :type crop_height: int (pixels)
            :param max_scale: maximum scale factor
            :type max_scale: int, default=1
            :param min_scale: minimum scale factor
            :type min_scale: int, default=0.9
            :param scale_step: scaling step
            :type scale_step: int, default=0.1
            :param step: step
            :type step: int, default=8
            :return: all_possible_crops_rect
            :rtype: list of CropRect
            """

        all_possible_crops_rects = self._get_all_possible_crops(
            image,
            crop_width,
            crop_height,
            max_scale=max_scale,
            min_scale=min_scale,
            scale_step=scale_step,
            step=step,
        )
        # Storing all feature maps in numpy stack
        feature_maps_image = np.dstack(extracted_feature_maps)
        start = time.time()
        for crop_rect in all_possible_crops_rects:
            crop_rect.score = self._score(feature_maps_image, crop_rect)

        end = time.time()
        # print("Time Spent in scoring Crop rectangles", end - start)
        self.extracted_feature_maps = extracted_feature_maps
        return all_possible_crops_rects

    def extract_candidate_crops(
        self, inputImage, crop_width, crop_height, feature_list
    ):
        """Public Function for crop_extractor module, Given input image and
        desired crop width and height: function returns list of all
        candidate crop rectangles scored taking into account input Image Feature list

        :param object: base class inheritance
        :type object: class:`Object`
        :param inputImage: input Image
        :type inputImage: opencv.Image
        :param crop_width: input crop width
        :type crop_width: Int
        :param crop_height: input crop height
        :type crop_height: Int
        :param feature_list: list of input feature maps to be used for scoring a crop rectangle
        :type feature_list: List of OpenCV numpy image type
        :return: List of extracted crop rectangle objects
        :rtype: list of crop_rect
        """
        extracted_candidate_crops = []
        extracted_feature_maps = []
        feature_names = []
        image = cv2.resize(
            inputImage,
            None,
            fx=1.0 / self.down_sample_factor,
            fy=1.0 / self.down_sample_factor,
        )

        for feature in feature_list:
            feature_names.append(type(feature).__name__)
            feature_map = feature.get_feature_map(image)
            extracted_feature_maps.append(feature_map)

        crop_width_small = int(crop_width / self.down_sample_factor)
        crop_height_small = int(crop_height / self.down_sample_factor)

        extracted_candidate_crops = self._extract_and_score_crop_rects(
            image,
            extracted_feature_maps,
            feature_names,
            crop_width_small,
            crop_height_small,
        )

        for crop_rect in extracted_candidate_crops:
            crop_rect.x = int(crop_rect.x * self.down_sample_factor)
            crop_rect.y = int(crop_rect.y * self.down_sample_factor)
            crop_rect.w = int(crop_rect.w * self.down_sample_factor)
            crop_rect.h = int(crop_rect.h * self.down_sample_factor)

        return extracted_candidate_crops
